#include <stdio.h>
#include <string.h>
#include "def.h"

#define LOCAL_DEBUG 0

extern int FreePage(struct PageHdr *PagePtr);
extern struct PageHdr *FetchPage(PAGENO page);
extern void strtolow(char *str);

static char *lookup_key;
static char *buf;
static int target_size;
static int tot_copied;

/******* A Stack for storing the page numbers ******/
struct stack_pages {
    PAGENO *pages; /* Array of pages */
    int top;       /* Current stack top */
    int cursize;   /* Current total allocated size of stack */
    int growBy;    /* Size by which stack is grown on over-running */
};
typedef struct stack_pages StackPages;

static StackPages *stack_init(int initial_size) {
    StackPages *sp = malloc(sizeof(StackPages));
    if (!sp)
        return NULL;

    sp->pages = malloc(sizeof(PAGENO) * initial_size);
    if (!sp->pages) {
        free(sp);
        return NULL;
    }

    sp->cursize = initial_size;
    sp->growBy = initial_size;
    sp->top = -1;

    return sp;
}

static void stack_destroy(StackPages *sp) {
    free(sp->pages);
    free(sp);

    return;
}

static int stack_empty(StackPages *sp) {
    if (sp->top < 0)
        return 1;

    return 0;
}

static int stack_full(StackPages *sp) {
    return sp->top >= (sp->cursize - 1);
}

void stack_push(StackPages *sp, PAGENO pgNum) {
    int new_size = 0;

    /* Our stack never over-runs, we simply realloc  and grow
 *      * the stack size */
    if (stack_full(sp)) {
        new_size = sp->cursize + sp->growBy;
        sp->pages = realloc(sp->pages, new_size);
        if (sp->pages == NULL) {
            fprintf(stderr, "Out of memory!!!\n");
            exit(-1);
        }
        sp->cursize = new_size;
    }

    sp->pages[++(sp->top)] = pgNum;

    return;
}

static PAGENO stack_pop(StackPages *sp) {
    if (stack_empty(sp)) {
        return -1;
    }

    return sp->pages[(sp->top)--];
}

/* display keys in a pge, used for debugging */
static inline void display_keys(struct PageHdr *p) {
#if LOCAL_DEBUG
    struct KeyRecord *records = p->KeyListPtr;
    int record_count = p->NumKeys;
    printf("---------------\n");
    printf("Record Count = %d\n", record_count);
    while (records != NULL) {
        printf("Key=%s\n", records->StoredKey);
        records = records->Next;
    }
    printf("---------------\n");
#endif

    return;
}

/* Add the keys in the leaf page to the circular buffer */
static int add_leaf_page_keys(const PAGENO leaf_page) {
    char *stored_key;
    int cmp_result;

    PAGENO page = leaf_page;
    struct PageHdr *p = FetchPage(page);
    assert(IsLeaf(p) && "not a leaf page");

    struct KeyRecord *records = p->KeyListPtr;
    int record_count = p->NumKeys;

    if (record_count <= 0)
        return 0;

    display_keys(p);

    do {
        stored_key = records->StoredKey;
        cmp_result = strcmp(stored_key, lookup_key);

        if (cmp_result >= 0)
            break; /* No need to traverse any further */

        strcpy(buf, stored_key);

        /* Total copied so far */
        tot_copied++;

        records = records->Next;
        record_count--;
    } while (record_count > 0);

    FreePage(p);
    return (tot_copied >= target_size);
}

/* Traverse down a b-tree, towards the appropriate root node */
static void vertical_search(PAGENO PageNo, int *done) {
    struct PageHdr *PagePtr = FetchPage(PageNo);
    PAGENO prev_page_num;
    StackPages *sp;
    NUMKEYS record_count;
    struct KeyRecord *records;
    int cmp_result;

    sp = stack_init(100); /* Stack of page numbers */

    if (IsLeaf(PagePtr)) {
        /* Aha, a leaf node. Add the keys */
        *done = add_leaf_page_keys(PageNo);
    } else {
        if (!sp) {
            fprintf(stderr, "Oops, out of memory!!!\n");
            exit(-1);
        }

        records = PagePtr->KeyListPtr;
        record_count = PagePtr->NumKeys;

        display_keys(PagePtr);
        do {
            cmp_result = strcmp(records->StoredKey, lookup_key);

            if (cmp_result >= 0) {
                /* Aha, found the right page. Traverse vertically down */
                vertical_search(records->PgNum, done);

                if (!(*done)) {
                    /* Not enough keys  in this node,
                     *  traverse horizontally backward ! */
                    prev_page_num = stack_pop(sp);
                    while (prev_page_num > 0 && !(*done)) {
                        vertical_search(prev_page_num, done);
                        prev_page_num = stack_pop(sp);
                    }
                }

                /* This node has been completely traversed at this point,
                 * break out and check if more keys need to be added */
                break;
            }

            if (records->Next == NULL) {
                /* A special case to find the rightmost child of this page
                 * Looks very ugly, :(. See if this can be refactored */
                stack_push(sp, records->PgNum);
                vertical_search(PagePtr->PtrToFinalRtgPg, done);
                if (!(*done)) {
                    /* Not enough keys  in this node,
                     *  traverse horizontally backward ! */
                    prev_page_num = stack_pop(sp);
                    while (prev_page_num > 0 && !(*done)) {
                        vertical_search(prev_page_num, done);
                        prev_page_num = stack_pop(sp);
                    }
                }

                /* This node has been completely traversed at this point,
                 * break out and check if more keys need to be added */
                break;
            } else {
                /* Store the left page number for later horizontal
                 *  backward traversal */
                stack_push(sp, records->PgNum);
                records = records->Next;
            }
            record_count--;
        } while (record_count > 0);
    }

    stack_destroy(sp);
    FreePage(PagePtr);
    return;
}

int get_leftbracket(char *key, char *result) {
    int done = 0;
    strtolow(key);
    buf = result;
    lookup_key = key;
    target_size = 1;
    tot_copied = 0;

    vertical_search(ROOT, &done);

    return tot_copied;
}
